@@ -1,7 +1,7 @@
 """This package contains several CircuitRunner implementations."""
 from __future__ import annotations
 
-__all__ = ['QuestRunner', 'SimulationRunner']
+__all__ = ['QuestRunner', 'SimulationRunner', 'gen_approximate_circuits']
 
 try:
     from bqskit.exec.runners.ibmq import IBMQRunner
@@ -10,4 +10,5 @@ except ImportError:
     pass
 
 from bqskit.exec.runners.quest import QuestRunner
+from bqskit.exec.runners.quest import gen_approximate_circuits
 from bqskit.exec.runners.sim import SimulationRunner
